/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.impl.annotationprocessor;

import static jp.objectfanatics.assertion.weaver.impl.core.WeaverUtils.createSimpleDescription;
import static jp.objectfanatics.assertion.weaver.impl.core.WeaverUtils.insertParameterAssertion;
import static jp.objectfanatics.assertion.weaver.impl.core.WeaverUtils.insertReturnValueAssertion;
import static jp.objectfanatics.assertion.weaver.impl.core.WeaverUtils.isAnnotatedByConstraintAnnotation;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getClassPool;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getParameterName;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getParameterType;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getReturnType;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.bytecode.annotation.Annotation;
import jp.objectfanatics.assertion.weaver.api.annotationprocessor.ConstraintAnnotationProcessor;
import jp.objectfanatics.assertion.weaver.api.core.exception.IllegalUseOfConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.api.core.exception.IllegalUseOfParameterConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.api.core.exception.IllegalUseOfReturnValueConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.impl.core.exception.IllegalUseOfParameterConstraintAnnotationExceptionImpl;
import jp.objectfanatics.assertion.weaver.impl.core.exception.IllegalUseOfReturnValueConstraintAnnotationExceptionImpl;
import jp.objectfanatics.commons.javassist.JavassistUtils.NoAvailableParameterNameInfoException;

/**
 * Abstract class for {@link ConstraintAnnotationProcessor}.
 * 
 * @author Makoto Sato
 */
abstract public class AbstractConstraintAnnotationProcessor implements ConstraintAnnotationProcessor {

	/**
	 * Return value identifier for javassist statement : "$_".
	 */
	public static final String RETURN_VALUE_IDENTIFIER = "$_";

	/**
	 * Target constraint annotation class.
	 */
	protected final Class<? extends java.lang.annotation.Annotation> targetConstraintAnnotation;

	/**
	 * Initializes a newly created {@link AbstractConstraintAnnotationProcessor} object with <code>targetConstraintAnnotation</code>.
	 * 
	 * @param targetConstraintAnnotation
	 */
	public AbstractConstraintAnnotationProcessor(Class<? extends java.lang.annotation.Annotation> targetConstraintAnnotation) {
		this.targetConstraintAnnotation = targetConstraintAnnotation;
	}

	@Override
	public final Class<? extends java.lang.annotation.Annotation> getTargetConstraintAnnotation() {
		return targetConstraintAnnotation;
	}

	@Override
	public final boolean processParameterConstraintAnnotation(CtBehavior ctBehavior, int paramIndex, Annotation annotation) throws IllegalUseOfConstraintAnnotationException {

		// return false if the annotation is not annotated by ConstraintAnnotation
		if (!isAnnotatedByConstraintAnnotation(getClassPool(ctBehavior), annotation)) return false;
		
		// get parameter type
		CtClass parameterType = getParameterType(ctBehavior, paramIndex);

		// assert that the specified parameter type is supported by this weaver
		assertParameterType(ctBehavior, paramIndex, annotation, parameterType);
		
		// call template method
		doWeaveParameterConstraintAssertion(ctBehavior, paramIndex, annotation);
		return true;
	}

	private void assertParameterType(CtBehavior ctBehavior, int paramIndex, Annotation annotation, CtClass parameterType) throws IllegalUseOfParameterConstraintAnnotationExceptionImpl {
		if (!isSupportedType(parameterType)) {
			String description = createSimpleDescription(ctBehavior, parameterType, annotation);
			throw new IllegalUseOfParameterConstraintAnnotationExceptionImpl(description, paramIndex, ctBehavior, annotation);
		}
	}

	// TODO: メソッド名要検討。
	/**
	 * Template method for {@link #processParameterConstraintAnnotation(CtBehavior, int, Annotation)}.
	 * 
	 * @param ctBehavior
	 *            a handle for dealing with the target constructor or method
	 * @param paramIndex
	 *            target parameter index
	 * @param annotation
	 *            target annotation
	 * @param paramType
	 *            target parameter type
	 * @param paramValueIdentifier
	 *            parameter value identifier for javassist statement
	 * @throws IllegalUseOfParameterConstraintAnnotationException
	 */
	abstract protected void doWeaveParameterConstraintAssertion(CtBehavior ctBehavior, int paramIndex, Annotation annotation) throws IllegalUseOfConstraintAnnotationException;

	@Override
	public final boolean processReturnValueConstraintAnnotation(CtMethod ctMethod, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException {

		// ignore if the annotation is not annotated by ConstraintAnnotation
		if (!isAnnotatedByConstraintAnnotation(getClassPool(ctMethod), annotation)) return false;
		
		// get return type
		CtClass returnType = getReturnType(ctMethod);
		
		// assert that the specified return type is supported by this weaver
		assertReturnType(ctMethod, annotation, returnType);
		
		// call template method
		doWeaveReturnValueConstraintAssertion(ctMethod, annotation);
		return true;
	}

	private void assertReturnType(CtMethod ctMethod, Annotation annotation, CtClass returnType) throws IllegalUseOfReturnValueConstraintAnnotationExceptionImpl {
		if (!isSupportedType(returnType)) {
			String description = createSimpleDescription(ctMethod, returnType, annotation);
			throw new IllegalUseOfReturnValueConstraintAnnotationExceptionImpl(description, ctMethod, annotation);
		}
	}

	/**
	 * Template method for {@link #processReturnValueConstraintAnnotation(CtMethod, Annotation)}
	 * 
	 * @param ctMethod
	 *            target method
	 * @param annotation
	 *            target annotation
	 * @throws IllegalUseOfReturnValueConstraintAnnotationException
	 */
	abstract protected void doWeaveReturnValueConstraintAssertion(CtMethod ctMethod, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException;

	// TODO: isSupported という名前のほうがよさそう。要検討。
	/**
	 * Returns whether the specified type is supported or not.
	 * <p/>
	 * This is a template method for {@link #processParameterConstraintAnnotation(CtBehavior, int, Annotation)} and {@link #processReturnValueConstraintAnnotation(CtMethod, Annotation)}.
	 * 
	 * @return
	 */
	abstract protected boolean isSupportedType(CtClass ctClass);

	/**
	 * Generates assertion statement string.
	 * 
	 * @param assertionErrorCondition
	 * @param message
	 * @return assertion statement string
	 */
	protected final String generateParameterAssrtionStatement(String assertionErrorCondition, String message) {
		return generateAssrtionStatement(assertionErrorCondition, message, IllegalArgumentException.class.getSimpleName());
	}

	/**
	 * Generates assertion statement string.
	 * 
	 * @param assertionErrorCondition
	 * @param message
	 * @return assertion statement string
	 */
	protected final String generateReturnValueAssrtionStatement(String assertionErrorCondition, String message) {
		return generateAssrtionStatement(assertionErrorCondition, message, IllegalStateException.class.getSimpleName());
	}

	/**
	 * Generates assertion statement string.
	 * 
	 * @param assertionErrorCondition
	 * @param message
	 * @return assertion statement string
	 */
	private String generateAssrtionStatement(String assertionErrorCondition, String message, String exceptionName) {
		return "if (" + assertionErrorCondition + ") throw new " + exceptionName + "(\"" + message + "\");";
	}

	protected final void weaveParameterConstraintAssertion(CtBehavior ctBehavior, String assertionErrorCondition, String message) {
		String assertionStatement = generateParameterAssrtionStatement(assertionErrorCondition, message);
		insertParameterAssertion(ctBehavior, assertionStatement);
	}

	protected final void weaveReturnValueConstraintAssertion(CtMethod ctMethod, String assertionErrorCondition, String message) {
		String assertionStatement = generateReturnValueAssrtionStatement(assertionErrorCondition, message);
		insertReturnValueAssertion(ctMethod, assertionStatement);
	}
	
	/**
	 * Returns a parameter information string for javassist statement.
	 * <p/>
	 * ex 1: <code>\"hoge\"(index=0)</code><br/>
	 * ex 2: <code>0</code><br/>
	 * 
	 * @param ctBehavior
	 *            a handle for dealing with the target constructor or method
	 * @param paramIndex
	 *            target parameter index
	 * @return a parameter information string
	 */
	protected final String createParamInfoStrForJavassist(CtBehavior ctBehavior, int paramIndex) {
		String paramInfoStr;
		try {
			// index is necessary for the user who does not have the source code.
			paramInfoStr = "\\\"" + getParameterName(ctBehavior, paramIndex) + "\\\"(index=" + paramIndex + ")";
		}
		catch (NoAvailableParameterNameInfoException e) {
			paramInfoStr = "" + paramIndex;
		}
		return paramInfoStr;
	}
	
	/**
	 * Returns a parameter information string.
	 * <p/>
	 * ex 1: <code>"hoge"(index=0)</code><br/>
	 * ex 2: <code>0</code><br/>
	 * 
	 * @param ctBehavior
	 *            a handle for dealing with the target constructor or method
	 * @param paramIndex
	 *            target parameter index
	 * @return a parameter information string
	 */
	protected final String createParamInfoStr(CtBehavior ctBehavior, int paramIndex) {
		String paramInfoStr;
		try {
			// index is necessary for the user who does not have the source code.
			paramInfoStr = "\"" + getParameterName(ctBehavior, paramIndex) + "\"(index=" + paramIndex + ")";
		}
		catch (NoAvailableParameterNameInfoException e) {
			paramInfoStr = "" + paramIndex;
		}
		return paramInfoStr;
	}
}
