/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.impl.annotationprocessor;

import java.text.MessageFormat;

import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.bytecode.annotation.Annotation;
import javassist.bytecode.annotation.DoubleMemberValue;
import jp.objectfanatics.assertion.constraints.NotEmpty;
import jp.objectfanatics.assertion.constraints.NotNull;
import jp.objectfanatics.assertion.constraints.Range;
import jp.objectfanatics.assertion.weaver.api.annotationprocessor.ConstraintAnnotationProcessor;
import jp.objectfanatics.assertion.weaver.api.core.exception.IllegalUseOfConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.api.core.exception.IllegalUseOfReturnValueConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.impl.annotationprocessor.numbercomparing.AbstractNumberComparingAssertionWeaver.ThresholdOutOfBoundsException;
import jp.objectfanatics.assertion.weaver.impl.core.exception.IllegalUseOfParameterConstraintAnnotationExceptionImpl;
import jp.objectfanatics.assertion.weaver.impl.core.exception.IllegalUseOfReturnValueConstraintAnnotationExceptionImpl;

/**
 * Implementation of {@link ConstraintAnnotationProcessor} for @{@link NotEmpty} annotation.
 * 
 * @author Makoto Sato
 */
public class RangeWeaverImpl extends AbstractConstraintAnnotationProcessor {

	/**
	 * for max check.
	 */
	private final MaxWeaverImpl maxAssertionWeaver = new MaxWeaverImpl();
	
	/**
	 * for min check.
	 */
	private final MinWeaverImpl minAssertionWeaver = new MinWeaverImpl();

	/**
	 * Constructor.
	 */
	public RangeWeaverImpl() {
		super(Range.class);
	}

	/**
	 * Returns true if, and only if, the parameter type is assignable to CharSequence, Collection, or array.
	 * 
	 * @param ctClass
	 *            type to be tested
	 * @return true if, and only if, the parameter type is assignable to CharSequence, Collection, or array.
	 */
	@Override
	protected boolean isSupportedType(@NotNull CtClass ctClass) {
		return maxAssertionWeaver.isSupportedType(ctClass) && minAssertionWeaver.isSupportedType(ctClass);
	}

	@Override
	protected void doWeaveParameterConstraintAssertion(CtBehavior ctBehavior, int paramIndex, Annotation annotation) throws IllegalUseOfConstraintAnnotationException {
		
		try {
			// weave prerequisite assertion
			weavePrerequisiteParameterConstraintAssertion(ctBehavior, paramIndex, annotation);
		}
		catch (NonFiniteThresholdException e) {
			throw new IllegalUseOfParameterConstraintAnnotationExceptionImpl(e.getDescription(), paramIndex, ctBehavior, annotation);
		}
		catch (ThresholdOutOfBoundsException e) {
			String description = e.getDescription();
			throw new IllegalUseOfParameterConstraintAnnotationExceptionImpl(description, paramIndex, ctBehavior, annotation);
		}
	}
	
	// TODO: 重複？
	private double getMinThreshold(CtBehavior ctBehavior, Annotation annotation) throws NonFiniteThresholdException {
		double threshold = DoubleMemberValue.class.cast(annotation.getMemberValue("min")).getValue();
		if (Double.isNaN(threshold) || Double.isInfinite(threshold)) {
			throw new NonFiniteThresholdException(ctBehavior, threshold);
		}
		return threshold;
	}
	// TODO: 重複？
	private double getMaxThreshold(CtBehavior ctBehavior, Annotation annotation) throws NonFiniteThresholdException {
		double threshold = DoubleMemberValue.class.cast(annotation.getMemberValue("max")).getValue();
		if (Double.isNaN(threshold) || Double.isInfinite(threshold)) {
			throw new NonFiniteThresholdException(ctBehavior, threshold);
		}
		return threshold;
	}

	// TODO: 重複？
	/**
	 * An error message pattern for using non-finite threshold error.
	 */
	private static final String WEAVING_ERROR_NON_FINITE_THRESHOLD_MESSAGE_PATTERN = "{0} is not supported by @{1} as a annotation parameter.";
	
	// TODO: 重複？
	private class NonFiniteThresholdException extends Exception {
		private final String description;

		public NonFiniteThresholdException(CtBehavior ctBehavior, double threshold) {
			this.description = createNonFiniteThresholdErrorMessage(threshold);
		}

		private String createNonFiniteThresholdErrorMessage(double threshold) {
			return MessageFormat.format(WEAVING_ERROR_NON_FINITE_THRESHOLD_MESSAGE_PATTERN, "" + threshold, targetConstraintAnnotation.getSimpleName());
		}

		public String getDescription() {
			return description;
		};
	}

	private void weavePrerequisiteParameterConstraintAssertion(CtBehavior ctBehavior, int paramIndex, Annotation annotation) throws IllegalUseOfConstraintAnnotationException, ThresholdOutOfBoundsException, NonFiniteThresholdException {
		double min = getMinThreshold(ctBehavior, annotation);
		double max = getMaxThreshold(ctBehavior, annotation);
		weaveRangeParameterConstraintAssertion(ctBehavior, paramIndex, min, max, annotation);
	}
	
	public void weaveRangeParameterConstraintAssertion(CtBehavior ctBehavior, int paramIndex, double min, double max, Annotation annotation) throws IllegalUseOfConstraintAnnotationException {
		minAssertionWeaver.weaveNumberComparingParameterConstraintAssertion(ctBehavior, paramIndex, min, annotation);
		maxAssertionWeaver.weaveNumberComparingParameterConstraintAssertion(ctBehavior, paramIndex, max, annotation);
	}

	@Override
	protected void doWeaveReturnValueConstraintAssertion(CtMethod ctMethod, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException {
		// weave prerequisite assertion
		try {
			weavePrerequisiteReturnValueConstraintAssertion(ctMethod, annotation);
		}
		catch (NonFiniteThresholdException e) {
			throw new IllegalUseOfReturnValueConstraintAnnotationExceptionImpl(e.getDescription(), ctMethod, annotation);
		}
	}

	private void weavePrerequisiteReturnValueConstraintAssertion(CtMethod ctMethod, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException, NonFiniteThresholdException {
		// FIXME: 単純にこのような呼び出し方はできない。annotation が深くかかわりすぎている。あるアノテーションに対してどのウィーバーを呼ぶという制御と、ある機能のウィービングは分離すべき。
		//       その上で、制御からウィービング機能への片方向依存をさせるのがたぶん正解。ということで、対応しましょう。
		double min = getMinThreshold(ctMethod, annotation);
		double max = getMaxThreshold(ctMethod, annotation);
		minAssertionWeaver.weaveNumberComparingReturnValueConstraintAssertion(ctMethod, min, annotation);
		maxAssertionWeaver.weaveNumberComparingReturnValueConstraintAssertion(ctMethod, max, annotation);
	}

	public void weaveRangeReturnValueConstraintAssertion(CtMethod ctMethod, double min, double max, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException {
		// TODO
	}
}
