/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.impl.annotationprocessor.numbercomparing;

import static jp.objectfanatics.commons.javassist.JavassistUtils.getParameterType;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrimitive;

import java.text.MessageFormat;

import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.bytecode.annotation.Annotation;
import javassist.bytecode.annotation.DoubleMemberValue;
import jp.objectfanatics.assertion.weaver.api.core.exception.IllegalUseOfConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.api.core.exception.IllegalUseOfReturnValueConstraintAnnotationException;
import jp.objectfanatics.assertion.weaver.impl.annotationprocessor.AbstractConstraintAnnotationProcessor;
import jp.objectfanatics.assertion.weaver.impl.annotationprocessor.NotNullAnnotationProcessor;
import jp.objectfanatics.assertion.weaver.impl.annotationprocessor.numbercomparing.assertionconditiongenerator.ComparableNumberAssertionConditionExpressionGenerators;
import jp.objectfanatics.assertion.weaver.impl.core.exception.IllegalUseOfParameterConstraintAnnotationExceptionImpl;
import jp.objectfanatics.assertion.weaver.impl.core.exception.IllegalUseOfReturnValueConstraintAnnotationExceptionImpl;
import jp.objectfanatics.commons.javassist.JavassistUtils;

// TODO: AssertionWeaver 系は全体的にかなりいじったので、全体を徹底的に整理しよう。 09.22.2009
/**
 * Abstract class for number comparing weavers.
 * 
 * @author Makoto Sato
 */
abstract public class AbstractNumberComparingAssertionWeaver extends AbstractConstraintAnnotationProcessor {

	/**
	 * An error message pattern for parameter constraint assertion error.
	 */
	private static final String ASSERTION_ERROR_MESSAGE_PATTERN = "{0}[=\"+{1}+\"] of {2} must not be {3} than {4}.";

	/**
	 * An error message pattern for using non-finite threshold error.
	 */
	private static final String WEAVING_ERROR_NON_FINITE_THRESHOLD_MESSAGE_PATTERN = "{0} is not supported by @{1} as a annotation parameter.";
	
	/**
	 * An error message pattern for threshold out of bounds error.
	 */
	private static final String WEAVING_ERROR_THRESHOLD_OUT_OF_BOUNDS_MESSAGE_PATTERN = "threshold for {0} must not be {1} than {2}. your setting is @{3}({4}).";

	/**
	 * For asserting non-primitive value is not null.
	 */
	private NotNullAnnotationProcessor notNullAssertionWeaver = new NotNullAnnotationProcessor();

	/**
	 * Comparative operator description.
	 * <p/>
	 * ex. <code>greater</code>
	 */
	private final String comparativeOperatorDescription;

	// TODO
	private ComparableNumberAssertionConditionExpressionGenerators comparableNumberAssertionConditionExpressionGenerators;

	/**
	 * Constructor.
	 * 
	 * @param targetConstraintAnnotation
	 */
	public AbstractNumberComparingAssertionWeaver(Class<? extends java.lang.annotation.Annotation> targetConstraintAnnotation, String comparativeOperator, String comparativeOperatorDescription) {
		// TODO: assertion を行うかどうかを要検討。ここだけでなく、ソース全体に対して。
		super(targetConstraintAnnotation);
		this.comparativeOperatorDescription = comparativeOperatorDescription;
		this.comparableNumberAssertionConditionExpressionGenerators = new ComparableNumberAssertionConditionExpressionGenerators(comparativeOperator, targetConstraintAnnotation);
	}

	@Override
	protected void doWeaveParameterConstraintAssertion(CtBehavior ctBehavior, int paramIndex, Annotation annotation) throws IllegalUseOfConstraintAnnotationException {

		try {
			// get threshold
			double threshold = getThreshold(ctBehavior, annotation);
			weaveNumberComparingParameterConstraintAssertion(ctBehavior, paramIndex, threshold, annotation);
		}
		catch (NonFiniteThresholdException e) {
			throw new IllegalUseOfParameterConstraintAnnotationExceptionImpl(e.getDescription(), paramIndex, ctBehavior, annotation);
		}
	}

	// TODO: 例外の情報を表示するだけのために annotation を受け取っているが、これは妥当なのか？ IllegalUseOfParameterConstraintAnnotationExceptionImpl から annotation 情報を省略した例外をここでは返して、上位で作り直すという手もある。要検討。
	public final void weaveNumberComparingParameterConstraintAssertion(CtBehavior ctBehavior, int paramIndex, double threshold, Annotation annotation) throws IllegalUseOfConstraintAnnotationException {
		try {
			String paramValueIdentifier = JavassistUtils.toParamValueIdentifier(ctBehavior, paramIndex);

			// get parameter type
			CtClass paramType = getParameterType(ctBehavior, paramIndex);

			// create assertion error condition
			String assertionErrorCondition = createAssertionErrorCondition(ctBehavior, paramType, paramValueIdentifier, threshold);

			// create assertion error message
			String message = createParameterConstraintAssertionErrorMessage(ctBehavior, paramIndex, paramValueIdentifier, threshold);

			// weave parameter constraint assertion
			weaveParameterConstraintAssertion(ctBehavior, assertionErrorCondition, message);

			// weave prerequisite assertion
			weavePrerequisiteParameterConstraintAssertionIfNeeded(ctBehavior, paramIndex, paramType, paramValueIdentifier);
		}
		catch (ThresholdOutOfBoundsException e) {
			String description = e.getDescription();
			throw new IllegalUseOfParameterConstraintAnnotationExceptionImpl(description, paramIndex, ctBehavior, annotation);
		}
	}

	// TODO: 重複？
	private double getThreshold(CtBehavior ctBehavior, Annotation annotation) throws NonFiniteThresholdException {
		double threshold = DoubleMemberValue.class.cast(annotation.getMemberValue("value")).getValue();
		if (Double.isNaN(threshold) || Double.isInfinite(threshold)) {
			throw new NonFiniteThresholdException(ctBehavior, threshold);
		}
		return threshold;
	}

	// TODO: 重複？
	private class NonFiniteThresholdException extends Exception {
		private final String description;

		public NonFiniteThresholdException(CtBehavior ctBehavior, double threshold) {
			this.description = createNonFiniteThresholdErrorMessage(threshold);
		}

		private String createNonFiniteThresholdErrorMessage(double threshold) {
			return MessageFormat.format(WEAVING_ERROR_NON_FINITE_THRESHOLD_MESSAGE_PATTERN, "" + threshold, targetConstraintAnnotation.getSimpleName());
		}

		public String getDescription() {
			return description;
		};
	}

	// ex. parameter "i"(index=0)[=-5] of Hoge.hoge(byte) is not supposed to be less than 10.
	private String createParameterConstraintAssertionErrorMessage(CtBehavior ctBehavior, int paramIndex, String paramValueIdentifier, double threshold) {
		String valueName = "parameter " + createParamInfoStrForJavassist(ctBehavior, paramIndex);
		String message = createAssertionErrorMessage(ctBehavior, paramValueIdentifier, threshold, valueName);
		return message;
	}

	private String createAssertionErrorMessage(CtBehavior ctBehavior, String paramValueIdentifier, double threshold, String valueName) {
		return MessageFormat.format(ASSERTION_ERROR_MESSAGE_PATTERN, valueName, paramValueIdentifier, ctBehavior.getLongName(), comparativeOperatorDescription, threshold);
	}

	private void weavePrerequisiteParameterConstraintAssertionIfNeeded(CtBehavior ctBehavior, int paramIndex, CtClass paramType, String paramValueIdentifier) throws IllegalUseOfConstraintAnnotationException {
		// delegates @NotNull assertion
		if (!isPrimitive(paramType)) notNullAssertionWeaver.weaveNotNullParameterConstraintAssertion(ctBehavior, paramIndex);
	}

	@Override
	protected void doWeaveReturnValueConstraintAssertion(CtMethod ctMethod, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException {

		try {
			
			// get threshold
			double threshold = getThreshold(ctMethod, annotation);
			
			weaveNumberComparingReturnValueConstraintAssertion(ctMethod, threshold, annotation);
		}
		catch (NonFiniteThresholdException e) {
			throw new IllegalUseOfReturnValueConstraintAnnotationExceptionImpl(e.getDescription(), ctMethod, annotation);
		}
	}

	public final void weaveNumberComparingReturnValueConstraintAssertion(CtMethod ctMethod, double threshold, Annotation annotation) throws IllegalUseOfReturnValueConstraintAnnotationException {
		try {
			CtClass returnType = JavassistUtils.getReturnType(ctMethod);

			// weave prerequisite assertion
			weavePrerequisiteReturnValueConstraintAssertionIfNeeded(ctMethod, annotation, returnType);

			// assertionErrorCondition の作成
			String assertionErrorCondition = createAssertionErrorCondition(ctMethod, returnType, RETURN_VALUE_IDENTIFIER, threshold);

			// メッセージの作成
			String message = createReturnValueConstraintAssertionErrorMessage(ctMethod, threshold);

			// 最終系
			weaveReturnValueConstraintAssertion(ctMethod, assertionErrorCondition, message);
		}
		catch (ThresholdOutOfBoundsException e) {
			String description = e.getDescription();
			throw new IllegalUseOfReturnValueConstraintAnnotationExceptionImpl(description, ctMethod, annotation);
		}
	}

	private String createReturnValueConstraintAssertionErrorMessage(CtMethod ctMethod, double threshold) {
		return MessageFormat.format(ASSERTION_ERROR_MESSAGE_PATTERN, "return value", RETURN_VALUE_IDENTIFIER, ctMethod.getLongName(), comparativeOperatorDescription, "" + threshold);
	}

	private void weavePrerequisiteReturnValueConstraintAssertionIfNeeded(CtMethod ctMethod, Annotation annotation, CtClass returnType) throws IllegalUseOfReturnValueConstraintAnnotationException {
		// delegates @NotNull assertion
		if (!isPrimitive(returnType)) notNullAssertionWeaver.weaveNotNullReturnValueConstraintAssertion(ctMethod);
	}

	// TODO: スコープ要検討。
	public static class ThresholdOutOfBoundsException extends Exception {
		private final String description;

		public ThresholdOutOfBoundsException(CtBehavior ctBehavior, double threshold, boolean isGreater, String limit, CtClass type, Class<? extends java.lang.annotation.Annotation> targetConstraintAnnotation) {
			this.description =  createThresholdOutOfBoundsErrorMessage(threshold, isGreater, limit, type, targetConstraintAnnotation);
		}

		// TODO: targetConstraintAnnotation の概念をこのクラスで扱うべきか、要検討。
		// TODO: 適切な引数名をつけよう。
		private String createThresholdOutOfBoundsErrorMessage(double threshold, boolean isGreater, String limit, CtClass type, Class<? extends java.lang.annotation.Annotation> targetConstraintAnnotation) {
			return MessageFormat.format(WEAVING_ERROR_THRESHOLD_OUT_OF_BOUNDS_MESSAGE_PATTERN, type.getName(), ((isGreater ? "greater" : "less")), "" + limit, targetConstraintAnnotation.getSimpleName(), "" + threshold);
		}

		public String getDescription() {
			return description;
		};
	}

	// TODO: 対応する型でない場合に、IllegalStateException を投げるのは妥当かどうか、要検討。
	private String createAssertionErrorCondition(CtBehavior ctBehavior, CtClass type, String valueIdentifier, double threshold) throws ThresholdOutOfBoundsException {
		return comparableNumberAssertionConditionExpressionGenerators.getAssertionErrorCondition(ctBehavior, type, valueIdentifier, threshold);
	}

	/**
	 * Returns true if, and only if, the specified type is byte, short, int, long, float, double, Byte, Short, Integer, Long, Float, Double, BigInteger, BigDecimal, AtomicInteger or AtomicLong.
	 * 
	 * @param type
	 *            limit type
	 * @return true if, and only if, the specified type is byte, short, int, long, float, double, Byte, Short, Integer, Long, Float, Double, BigInteger, BigDecimal, AtomicInteger or AtomicLong.
	 */
	@Override
	public boolean isSupportedType(CtClass type) {
		return comparableNumberAssertionConditionExpressionGenerators.isSupported(type);
	}
}
