/*
 * Copyright (C) 2005-2009 Makoto Sato. All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package jp.objectfanatics.assertion.weaver.impl.core;

import static jp.objectfanatics.commons.javassist.JavassistUtils.getAnnotations;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getAssignableTypesExceptItself;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getParameterAnnotations;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getParameterNames;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getParameterTypesWithoutContextParameter;
import static jp.objectfanatics.commons.javassist.JavassistUtils.getSimpleName;
import static jp.objectfanatics.commons.javassist.JavassistUtils.insertAfter;
import static jp.objectfanatics.commons.javassist.JavassistUtils.insertBefore;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAbstract;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isAnnotatedWithMetaAnnotation;
import static jp.objectfanatics.commons.javassist.JavassistUtils.isPrivate;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javassist.ClassPool;
import javassist.CtBehavior;
import javassist.CtClass;
import javassist.CtMethod;
import javassist.bytecode.AnnotationsAttribute;
import javassist.bytecode.ClassFile;
import javassist.bytecode.ConstPool;
import javassist.bytecode.annotation.Annotation;
import jp.objectfanatics.assertion.constraints.ConstraintAnnotation;
import jp.objectfanatics.assertion.constraints.Woven;
import jp.objectfanatics.assertion.weaver.api.annotationprocessor.ConstraintAnnotationProcessor;
import jp.objectfanatics.commons.javassist.JavassistUtils;
import jp.objectfanatics.commons.javassist.JavassistUtils.NoAvailableParameterNameInfoException;

public class WeaverUtils {
    
    // TODO: このようなメソッドを作成しても、大本の insertAfter を呼ぶコードがあったら全く意味がない。
    //       JDepend の maven plugin で対応できるのであれば、それで対応しよう。
    //       もしくは、指定されたメソッドや behavior への呼び出しをチェックするプログラムを別 artifact として作成しよう。
    //       そのアーティファクト内に、Mojoも作ればOK。 
    public static void insertReturnValueAssertion(CtMethod ctMethod, String assertionStatement) {
        if (!isAbstract(ctMethod)) insertAfter(ctMethod, assertionStatement);
    }
    
    public static void insertParameterAssertion(CtBehavior ctBehavior, String assertionStatement) {
        if (!isAbstract(ctBehavior)) insertBefore(ctBehavior, assertionStatement);
    }
	
	/**
	 * Returns simple description like '@AnAnnotation does not support "int".'.
	 * @param ctBehavior {@link CtBehavior}
	 * @param targetType 
	 * @param annotation
	 * @return
	 */
	public static String createSimpleDescription(CtBehavior ctBehavior, CtClass targetType, Annotation annotation) {
		return "@" + getAnnotationSimpleName(ctBehavior, annotation) + " does not support \"" + targetType.getSimpleName() + "\".";
	}
	
	/**
	 * Short cut method : JavassistUtils.getSimpleName(ctBehavior.getDeclaringClass().getClassPool(), annotation)
	 * 
	 * @param ctBehavior
	 *            {@link CtBehavior}
	 * @param annotation
	 *            {@link Annotation}
	 * @return simple name of <code>annotation</code>
	 */
	static String getAnnotationSimpleName(CtBehavior ctBehavior, Annotation annotation) {
		return getSimpleName(ctBehavior.getDeclaringClass().getClassPool(), annotation);
	}
	
	/**
	 * Weaves {@link Woven} annotation into given class file bytes.
	 * 
	 * @param ctClass
	 *            a handle for dealing with the target class file
	 * @return woven class file bytes
	 * @throws IOException
	 */
	static void weaveWovenFlag(CtClass ctClass) {
		ClassFile cf = ctClass.getClassFile();
		ConstPool cp = cf.getConstPool();
		AnnotationsAttribute attr = (AnnotationsAttribute) cf.getAttribute(AnnotationsAttribute.visibleTag);
		if (attr == null) {
			attr = new AnnotationsAttribute(cp, AnnotationsAttribute.visibleTag);
		}
		Annotation a = new Annotation(Woven.class.getName(), cp);
		attr.addAnnotation(a);
		cf.addAttribute(attr);
		cf.setVersionToJava5();
	}
	
	/**
	 * Returns true if, and only if, annotation is annotated by {@link ConstraintAnnotation}.
	 * 
	 * @param annotation
	 *            target
	 * @return true if, and only if, annotation is annotated by {@link ConstraintAnnotation}
	 */
	public static boolean isAnnotatedByConstraintAnnotation(ClassPool classPool, Annotation annotation) {
		return isAnnotatedWithMetaAnnotation(classPool, annotation, ConstraintAnnotation.class.getName());
	}
	
	/**
	 * Returns target {@link ConstraintAnnotationProcessor} for the specified {@link Annotation}.
	 * 
	 * @param annotation
	 * @return target {@link ConstraintAnnotationProcessor} for the specified {@link Annotation}.
	 * @throws TargetConstraintAnnotationAssertionWeaverNotFoundException
	 */
	static ConstraintAnnotationProcessor getTargetConstraintAnnotationAssertionWeaver(Annotation annotation, ConstraintAnnotationProcessor[] constraintAssertionWeavers) throws TargetConstraintAnnotationAssertionWeaverNotFoundException {
		for (ConstraintAnnotationProcessor constraintAnnotationAssertionWeaver : constraintAssertionWeavers) {
			if (annotation.getTypeName().equals(constraintAnnotationAssertionWeaver.getTargetConstraintAnnotation().getName())) {
				return constraintAnnotationAssertionWeaver;
			}
		}
		throw new TargetConstraintAnnotationAssertionWeaverNotFoundException();
//		throw new IllegalStateException(annotation.getTypeName() + " is annotated by " + ConstraintAnnotation.class.getName() + ", but no handler found.");
	}
	static class TargetConstraintAnnotationAssertionWeaverNotFoundException extends Exception {}
	
	static Set<CtMethod> getAncestorMethods(CtMethod ctBehavior) {
		if (isPrivate(ctBehavior)) return new HashSet<CtMethod>();
		Set<CtMethod> ancestorMethods = new HashSet<CtMethod>();
		for (CtClass ancestorClassesOrInterface : getAssignableTypesExceptItself(ctBehavior)) {
			for (CtMethod candidate : ancestorClassesOrInterface.getMethods()) {
				if (isPrivate(candidate)) continue;
				if (!ctBehavior.getName().equals(candidate.getName())) continue;
				if (hasSameMethodParameterTypes(ctBehavior, candidate)) ancestorMethods.add(candidate);
			}
		}
		return ancestorMethods;
	}
	
	static boolean hasSameMethodParameterTypes(CtMethod targetMethod, CtMethod ancestorCandidateMethod) {
		CtClass[] targetMethodParameterTypes = getParameterTypesWithoutContextParameter(targetMethod);
		CtClass[] ancestorCandidateMethodParameterTypes = getParameterTypesWithoutContextParameter(ancestorCandidateMethod);
		if (targetMethodParameterTypes.length != ancestorCandidateMethodParameterTypes.length) return false;
		for (int i = 0; i < targetMethodParameterTypes.length; i++) {
			String targetMethodParameterTypeName = targetMethodParameterTypes[i].getName();
			String ancestorCandidateMethodParameterTypeName = ancestorCandidateMethodParameterTypes[i].getName();
			if (!targetMethodParameterTypeName.equals(ancestorCandidateMethodParameterTypeName)) {
				return false;
			}
		}
		return true;
	}
	
	static Annotation[] getParameterConstraintAnnotations(ClassPool classPool, CtBehavior ctBehavior, int paramIndex) {
		List<Annotation> parameterConstraintAnnotations = new ArrayList<Annotation>();
		for (Annotation parameterAnnotation : getParameterAnnotations(ctBehavior, paramIndex)) {
			if (isAnnotatedByConstraintAnnotation(classPool, parameterAnnotation)) parameterConstraintAnnotations.add(parameterAnnotation);
		}
		return parameterConstraintAnnotations.toArray(new Annotation[parameterConstraintAnnotations.size()]);
	}
	
	static Annotation[] getReturnValueConstraintAnnotations(CtBehavior ctBehavior) {
		List<Annotation> retrunValueConstraintAnnotations = new ArrayList<Annotation>();
		for (Annotation annotation : getAnnotations(ctBehavior)) {
			if (isAnnotatedByConstraintAnnotation(JavassistUtils.getClassPool(ctBehavior), annotation)) retrunValueConstraintAnnotations.add(annotation);
		}
		return retrunValueConstraintAnnotations.toArray(new Annotation[retrunValueConstraintAnnotations.size()]);
	}
	
	public static String[] getParameterNamesEvenIfNoAvailableParameterNameInfo(CtBehavior ctMethod) {
		try {
			return getParameterNames(ctMethod);
		}
		catch (NoAvailableParameterNameInfoException e) {}
		String[] parameterNames = new String[getParameterTypesWithoutContextParameter(ctMethod).length];
		for (int i = 0; i < parameterNames.length; i++) {
			parameterNames[i] = "" + i + "(zero base)";
		}
		return parameterNames;
	}
	
	// てすとちう
	/**
	 * Returns true if, and only if, assertion functions are already woven into the class file bytes.
	 * 
	 * @param ctClass
	 *            a handle for dealing with the target class file
	 * @return true if, and only if, assertion functions are already woven into the class file bytes
	 */
	static boolean isAlreadyWoven(CtClass ctClass) {
		return JavassistUtils.isAnnotated(ctClass, Woven.class.getName());
	}
	
	// --------------------------------------------------------------------------------------------
	// Unclassified
	// --------------------------------------------------------------------------------------------
	// Checked(9.19.2009), Tested(9.19.2009)
}
